@@ -3,16 +3,28 @@
 '''
 A library to access Tryton's models like a client.
 '''
-__version__ = "2.2.4"
+__version__ = "2.4.0"
 __all__ = ['Model', 'Wizard']
-from types import NoneType
+import sys
+try:
+    import cdecimal
+    # Use cdecimal globally
+    if 'decimal' not in sys.modules:
+        sys.modules['decimal'] = cdecimal
+except ImportError:
+    import decimal
+    sys.modules['cdecimal'] = decimal
 import threading
 import datetime
 from decimal import Decimal
+from types import NoneType
+
 import proteus.config
+from proteus.pyson import PYSONDecoder
 
 _MODELS = threading.local()
 
+
 class _EvalEnvironment(dict):
     'Dictionary for evaluation'
     def __init__(self, parent):
@@ -57,12 +69,12 @@ class FieldDescriptor(object):
         self.__doc__ = definition['string']
 
     def __get__(self, instance, owner):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         return instance._values.get(self.name, self.default)
 
     def __set__(self, instance, value):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         previous = getattr(instance, self.name)
         instance._values[self.name] = value
@@ -94,30 +106,31 @@ class BinaryDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, (basestring, buffer)) or value in (None, False)
+        assert (isinstance(value, (basestring, buffer))
+            or value in (None, False))
         super(BinaryDescriptor, self).__set__(instance, value or '')
 
 
 class IntegerDescriptor(FieldDescriptor):
-    default = 0
 
     def __set__(self, instance, value):
-        assert isinstance(value, (int, long))
+        assert isinstance(value, (int, long, NoneType))
         super(IntegerDescriptor, self).__set__(instance, value)
 
 
 class FloatDescriptor(FieldDescriptor):
-    default = 0.0
 
     def __set__(self, instance, value):
-        super(FloatDescriptor, self).__set__(instance, float(value))
+        assert isinstance(value, (int, long, float, Decimal, NoneType))
+        if value is not None:
+            value = float(value)
+        super(FloatDescriptor, self).__set__(instance, value)
 
 
 class NumericDescriptor(FieldDescriptor):
-    default = Decimal('0.0')
 
     def __set__(self, instance, value):
-        assert isinstance(value, Decimal)
+        assert isinstance(value, (NoneType, Decimal))
         # TODO add digits validation
         super(NumericDescriptor, self).__set__(instance, value)
 
@@ -162,6 +175,12 @@ class DateTimeDescriptor(FieldDescriptor):
         super(DateTimeDescriptor, self).__set__(instance, value)
 
 
+class TimeDescriptor(FieldDescriptor):
+    def __set__(self, instance, value):
+        assert isinstance(value, datetime.time) or value in (None, False)
+        super(TimeDescriptor, self).__set__(instance, value)
+
+
 class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         relation = Model.get(self.definition['relation'], instance._config)
@@ -170,8 +189,7 @@ class Many2OneDescriptor(FieldDescriptor):
             value = relation(value)
         elif not value:
             value = None
-        if self.name in instance._values:
-            instance._values[self.name] = value
+        instance._values[self.name] = value
         return value
 
     def __set__(self, instance, value):
@@ -304,7 +322,7 @@ class MetaModelFactory(object):
         'text': CharDescriptor,
         'sha': CharDescriptor,
         'binary': BinaryDescriptor,
-        'selection': CharDescriptor, # TODO implement its own descriptor
+        'selection': CharDescriptor,  # TODO implement its own descriptor
         'integer': IntegerDescriptor,
         'biginteger': IntegerDescriptor,
         'float': FloatDescriptor,
@@ -313,6 +331,7 @@ class MetaModelFactory(object):
         'reference': ReferenceDescriptor,
         'date': DateDescriptor,
         'datetime': DateTimeDescriptor,
+        'time': TimeDescriptor,
         'many2one': Many2OneDescriptor,
         'one2many': One2ManyDescriptor,
         'many2many': Many2ManyDescriptor,
@@ -342,6 +361,7 @@ class MetaModelFactory(object):
         models_key = 'c%s' % id(self.config)
         if not hasattr(_MODELS, models_key):
             setattr(_MODELS, models_key, {})
+
         class MetaModel(type):
             'Meta class for Model'
             def __new__(mcs, name, bases, dict):
@@ -493,11 +513,11 @@ class Model(object):
         self.__id = id or Model.__counter
         if self.__id < 0:
             Model.__counter -= 1
-        self._values = {} # store the values of fields
-        self._changed = set() # store the changed fields
-        self._parent = None # store the parent record
-        self._parent_field_name = '' # store the field name in parent record
-        self._parent_name = '' # store the field name to parent record
+        self._values = {}  # store the values of fields
+        self._changed = set()  # store the changed fields
+        self._parent = None  # store the parent record
+        self._parent_field_name = ''  # store the field name in parent record
+        self._parent_name = ''  # store the field name to parent record
         if self.id < 0 and _default:
             self._default_get()
 
@@ -521,9 +541,6 @@ class Model(object):
     @classmethod
     def get(cls, name, config=None):
         'Get a class for the named Model'
-        if isinstance(name, unicode):
-            name = name.encode('utf-8')
-
         class Spam(Model):
             __metaclass__ = MetaModelFactory(name, config=config)()
         return Spam
@@ -583,7 +600,7 @@ class Model(object):
         'Save the record'
         context = self._config.context
         if self.id < 0:
-            values = self._get_values()
+            values = self._get_values(fields=self._changed)
             self.__id = self._proxy.create(values, context)
         else:
             if not self._changed:
@@ -628,12 +645,15 @@ class Model(object):
 
     def _read(self, name):
         'Read field'
-        fields = []
-        if not self._values:
+        fields = [name]
+        if name in self._values:
+            return
+        loading = self._fields[name]['loading']
+        if loading == 'eager':
             fields = [x for x, y in self._fields.iteritems()
-                    if y['type'] not in ('one2many', 'many2many', 'binary')]
+                    if y['loading'] == 'eager']
+        if not self._fields:
             fields.append('_timestamp')
-        fields.append(name)
         self._values.update(self._proxy.read(self.id, fields,
             self._config.context))
         for field in fields:
@@ -651,6 +671,8 @@ class Model(object):
 
     def _default_set(self, values):
         for field, value in values.iteritems():
+            if '.' in field:
+                continue
             definition = self._fields[field]
             if definition['type'] in ('one2many', 'many2many'):
                 if value and len(value) and isinstance(value[0], (int, long)):
@@ -673,13 +695,17 @@ class Model(object):
         values['id'] = self.id
         return values
 
-    def _on_change_args(self, args):
-        res = {}
+    def _get_on_change_value(self):
         values = self._get_eval()
         del values['id']
         for field, definition in self._fields.iteritems():
             if definition['type'] in ('one2many', 'many2many'):
                 values[field] = [x._get_eval() for x in getattr(self, field)]
+        return values
+
+    def _on_change_args(self, args):
+        res = {}
+        values = self._get_on_change_value()
         if self._parent:
             values['_parent_%s' % self._parent_name] = \
                     _EvalEnvironment(self._parent)
@@ -733,8 +759,6 @@ class Model(object):
 
     def _on_change(self, name):
         'Call on_change for field'
-        # Import locally to not break installation
-        from proteus.pyson import PYSONDecoder
         definition = self._fields[name]
         if definition.get('on_change'):
             if isinstance(definition['on_change'], basestring):
@@ -753,6 +777,9 @@ class Model(object):
                 self._on_change_set(field, value)
             for field, value in later.iteritems():
                 self._on_change_set(field, value)
+        values = {}
+        to_change = set()
+        later = set()
         for field, definition in self._fields.iteritems():
             if not definition.get('on_change_with'):
                 continue
@@ -760,11 +787,23 @@ class Model(object):
                 continue
             if field == name:
                 continue
-            args = self._on_change_args(definition['on_change_with'])
+            if to_change & set(definition['on_change_with']):
+                later.add(field)
+            to_change.add(field)
+            values.update(self._on_change_args(definition['on_change_with']))
+        if to_change:
             context = self._config.context
-            res = getattr(self._proxy, 'on_change_with_%s' % field)(args,
+            result = getattr(self._proxy, 'on_change_with')(list(to_change),
+                values, context)
+            for field, value in result.items():
+                self._on_change_set(field, value)
+        for field in later:
+            definition = self._fields[field]
+            values = self._on_change_args(definition['on_change_with'])
+            context = self._config.context
+            result = getattr(self._proxy, 'on_change_with_%s' % field)(values,
                     context)
-            self._on_change_set(field, res)
+            self._on_change_set(field, result)
         if self._parent:
             self._parent._changed.add(self._parent_field_name)
             self._parent._on_change(self._parent_field_name)
@@ -778,62 +817,53 @@ class Wizard(object):
             assert len(set(type(x) for x in models)) == 1
         super(Wizard, self).__init__()
         self.name = name
-        self.state = None
-        self.states = ['init']
         self.form = None
+        self.form_state = None
         self._config = config or proteus.config.get_config()
         self._context = context or {}
         self._proxy = self._config.get_proxy(name, type='wizard')
-        self.id = self._proxy.create(self._config.context)
-        if models:
-            self.datas = {
-                    'model': models[0].__class__.__name__,
-                    'id': models[0].id,
-                    'ids': [model.id for model in models]
-                    }
-        else:
-            self.datas = {}
-        self.execute('init')
+        result = self._proxy.create(self._config.context)
+        self.session_id, self.start_state, self.end_state = result
+        self.states = [self.start_state]
+        self.models = models
+        self.execute(self.start_state)
 
     def execute(self, state):
         assert state in self.states
 
-        if 'form' not in self.datas:
-            self.datas['form'] = {}
-
-        if self.form:
-            self.datas['form'].update(self.form._get_values())
-
         self.state = state
         while self.state != 'end':
             ctx = self._context.copy()
             ctx.update(self._config.context)
-            ctx['active_id'] = self.datas.get('id')
-            ctx['active_ids'] = self.datas.get('ids')
+            if self.models:
+                ctx['active_id'] = self.models[0].id
+                ctx['active_ids'] = [model.id for model in self.models]
+                ctx['active_model'] = self.models[0].__class__.__name__
+            else:
+                ctx['active_id'] = None
+                ctx['active_ids'] = None
+                ctx['active_model'] = None
 
-            res = self._proxy.execute(self.id, self.datas, self.state, ctx)
-            if not res:
-                break
+            if self.form:
+                data = {self.form_state: self.form._get_on_change_value()}
+            else:
+                data = {}
 
-            if 'datas' in res:
-                self.datas['form'] = res['datas']
-            elif res['type'] == 'form':
-                self.datas['form'] = {}
+            result = self._proxy.execute(self.session_id, data, self.state,
+                ctx)
 
-            if res['type'] == 'form':
-                self.states = [x[0] for x in res['state']]
-                # XXX set context
-                self.form = Model.get(res['object'])()
-                self.form._default_set(self.datas['form'])
+            if 'view' in result:
+                view = result['view']
+                self.form = Model.get(view['fields_view']['model'])()
+                self.form._default_set(view['defaults'])
+                self.states = [b['state'] for b in view['buttons']]
+                self.form_state = view['state']
                 return
-            elif res['type'] == 'action':
-                # TODO run action
-                self.state = res['state']
-            elif res['type'] == 'print':
-                # TODO run print
-                self.state = res['state']
-            elif res['type'] == 'state':
-                self.state = res['state']
+            else:
+                self.state = self.end_state
 
-        if self.state == 'end':
-            self._proxy.delete(self.id, self._config.context)
+            if 'actions' in result:
+                pass  # TODO
+
+        if self.state == self.end_state:
+            self._proxy.delete(self.session_id, self._config.context)
